//=======================================================================================
// ____          ____    __    ______     __________   __      __       __        __
// \    \       |    |  |  |  |   _   \  |___    ___| |  |    |  |     /  \      |  |
//  \    \      |    |  |  |  |  |_)   |     |  |     |  |    |  |    /    \     |  |
//   \    \     |    |  |  |  |   _   /      |  |     |  |    |  |   /  /\  \    |  |
//    \    \    |    |  |  |  |  | \  \      |  |     |   \__/   |  /  ____  \   |  |____
//     \    \   |    |  |__|  |__|  \__\     |__|      \________/  /__/    \__\  |_______|
//      \    \  |    |   ________________________________________________________________
//       \    \ |    |  |  ______________________________________________________________|
//        \    \|    |  |  |         __          __     __     __     ______      _______
//         \         |  |  |_____   |  |        |  |   |  |   |  |   |   _  \    /  _____)
//          \        |  |   _____|  |  |        |  |   |  |   |  |   |  | \  \   \_______
//           \       |  |  |        |  |_____   |   \_/   |   |  |   |  |_/  /    _____  |
//            \ _____|  |__|        |________|   \_______/    |__|   |______/    (_______/
//
//  This file is part of VirtualFluids. VirtualFluids is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  VirtualFluids is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with VirtualFluids (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file QVelDevPlainBB27.cu
//! \ingroup BoundaryConditions
//! \author Martin Schoenherr
//=======================================================================================
/* Device code */
#include "LBM/LB.h"
#include "LBM/D3Q27.h"
#include <lbm/constants/NumericConstants.h>

using namespace vf::lbm::constant;
//////////////////////////////////////////////////////////////////////////////
extern "C" __global__ void QVelDevPlainBB27(
    real* velocityX,
    real* velocityY,
    real* velocityZ,
    real* distributions,
    int* subgridDistanceIndices,
    real* subgridDistances,
    uint numberOfSubgridIndices,
    int numberOfBCnodes,
    uint* neighborX,
    uint* neighborY,
    uint* neighborZ,
    uint numberOfLBnodes,
    bool isEvenTimestep)
{
    //////////////////////////////////////////////////////////////////////////
    //! The velocity boundary condition is executed in the following steps
    //!
    ////////////////////////////////////////////////////////////////////////////////
    //! - Get node index coordinates from thredIdx, blockIdx, blockDim and gridDim.
    //!
    const unsigned  x = threadIdx.x;  // global x-index
   const unsigned  y = blockIdx.x;   // global y-index
   const unsigned  z = blockIdx.y;   // global z-index

   const unsigned nx = blockDim.x;
   const unsigned ny = gridDim.x;

   const unsigned k = nx*(ny*z + y) + x;

   //////////////////////////////////////////////////////////////////////////
   // run for all indices in size of boundary condition (numberOfBCnodes)
   if(k < numberOfBCnodes)
   {
       //////////////////////////////////////////////////////////////////////////
       //! - Read distributions: style of reading and writing the distributions from/to stored arrays dependent on timestep is based on the esoteric twist algorithm \ref
       //! <a href="https://doi.org/10.3390/computation5020019"><b>[ M. Geier et al. (2017), DOI:10.3390/computation5020019 ]</b></a>
       //!
       Distributions27 dist;
       if (isEvenTimestep)
       {
          dist.f[dirE   ] = &distributions[dirE   *numberOfLBnodes];
          dist.f[dirW   ] = &distributions[dirW   *numberOfLBnodes];
          dist.f[dirN   ] = &distributions[dirN   *numberOfLBnodes];
          dist.f[dirS   ] = &distributions[dirS   *numberOfLBnodes];
          dist.f[dirT   ] = &distributions[dirT   *numberOfLBnodes];
          dist.f[dirB   ] = &distributions[dirB   *numberOfLBnodes];
          dist.f[dirNE  ] = &distributions[dirNE  *numberOfLBnodes];
          dist.f[dirSW  ] = &distributions[dirSW  *numberOfLBnodes];
          dist.f[dirSE  ] = &distributions[dirSE  *numberOfLBnodes];
          dist.f[dirNW  ] = &distributions[dirNW  *numberOfLBnodes];
          dist.f[dirTE  ] = &distributions[dirTE  *numberOfLBnodes];
          dist.f[dirBW  ] = &distributions[dirBW  *numberOfLBnodes];
          dist.f[dirBE  ] = &distributions[dirBE  *numberOfLBnodes];
          dist.f[dirTW  ] = &distributions[dirTW  *numberOfLBnodes];
          dist.f[dirTN  ] = &distributions[dirTN  *numberOfLBnodes];
          dist.f[dirBS  ] = &distributions[dirBS  *numberOfLBnodes];
          dist.f[dirBN  ] = &distributions[dirBN  *numberOfLBnodes];
          dist.f[dirTS  ] = &distributions[dirTS  *numberOfLBnodes];
          dist.f[dirREST] = &distributions[dirREST*numberOfLBnodes];
          dist.f[dirTNE ] = &distributions[dirTNE *numberOfLBnodes];
          dist.f[dirTSW ] = &distributions[dirTSW *numberOfLBnodes];
          dist.f[dirTSE ] = &distributions[dirTSE *numberOfLBnodes];
          dist.f[dirTNW ] = &distributions[dirTNW *numberOfLBnodes];
          dist.f[dirBNE ] = &distributions[dirBNE *numberOfLBnodes];
          dist.f[dirBSW ] = &distributions[dirBSW *numberOfLBnodes];
          dist.f[dirBSE ] = &distributions[dirBSE *numberOfLBnodes];
          dist.f[dirBNW ] = &distributions[dirBNW *numberOfLBnodes];
       }
       else
       {
          dist.f[dirW   ] = &distributions[dirE   *numberOfLBnodes];
          dist.f[dirE   ] = &distributions[dirW   *numberOfLBnodes];
          dist.f[dirS   ] = &distributions[dirN   *numberOfLBnodes];
          dist.f[dirN   ] = &distributions[dirS   *numberOfLBnodes];
          dist.f[dirB   ] = &distributions[dirT   *numberOfLBnodes];
          dist.f[dirT   ] = &distributions[dirB   *numberOfLBnodes];
          dist.f[dirSW  ] = &distributions[dirNE  *numberOfLBnodes];
          dist.f[dirNE  ] = &distributions[dirSW  *numberOfLBnodes];
          dist.f[dirNW  ] = &distributions[dirSE  *numberOfLBnodes];
          dist.f[dirSE  ] = &distributions[dirNW  *numberOfLBnodes];
          dist.f[dirBW  ] = &distributions[dirTE  *numberOfLBnodes];
          dist.f[dirTE  ] = &distributions[dirBW  *numberOfLBnodes];
          dist.f[dirTW  ] = &distributions[dirBE  *numberOfLBnodes];
          dist.f[dirBE  ] = &distributions[dirTW  *numberOfLBnodes];
          dist.f[dirBS  ] = &distributions[dirTN  *numberOfLBnodes];
          dist.f[dirTN  ] = &distributions[dirBS  *numberOfLBnodes];
          dist.f[dirTS  ] = &distributions[dirBN  *numberOfLBnodes];
          dist.f[dirBN  ] = &distributions[dirTS  *numberOfLBnodes];
          dist.f[dirREST] = &distributions[dirREST*numberOfLBnodes];
          dist.f[dirTNE ] = &distributions[dirBSW *numberOfLBnodes];
          dist.f[dirTSW ] = &distributions[dirBNE *numberOfLBnodes];
          dist.f[dirTSE ] = &distributions[dirBNW *numberOfLBnodes];
          dist.f[dirTNW ] = &distributions[dirBSE *numberOfLBnodes];
          dist.f[dirBNE ] = &distributions[dirTSW *numberOfLBnodes];
          dist.f[dirBSW ] = &distributions[dirTNE *numberOfLBnodes];
          dist.f[dirBSE ] = &distributions[dirTNW *numberOfLBnodes];
          dist.f[dirBNW ] = &distributions[dirTSE *numberOfLBnodes];
       }


      ////////////////////////////////////////////////////////////////////////////////
      //! - Set local velocities
      //!
      real VeloX = velocityX[k];
      real VeloY = velocityY[k];
      real VeloZ = velocityZ[k];
      ////////////////////////////////////////////////////////////////////////////////
      //! - Set local subgrid distances (q's)
      //!
      real   *q_dirE,   *q_dirW,   *q_dirN,   *q_dirS,   *q_dirT,   *q_dirB,
             *q_dirNE,  *q_dirSW,  *q_dirSE,  *q_dirNW,  *q_dirTE,  *q_dirBW,
             *q_dirBE,  *q_dirTW,  *q_dirTN,  *q_dirBS,  *q_dirBN,  *q_dirTS,
             *q_dirTNE, *q_dirTSW, *q_dirTSE, *q_dirTNW, *q_dirBNE, *q_dirBSW,
             *q_dirBSE, *q_dirBNW;
      q_dirE   = &subgridDistances[dirE   *numberOfSubgridIndices];
      q_dirW   = &subgridDistances[dirW   *numberOfSubgridIndices];
      q_dirN   = &subgridDistances[dirN   *numberOfSubgridIndices];
      q_dirS   = &subgridDistances[dirS   *numberOfSubgridIndices];
      q_dirT   = &subgridDistances[dirT   *numberOfSubgridIndices];
      q_dirB   = &subgridDistances[dirB   *numberOfSubgridIndices];
      q_dirNE  = &subgridDistances[dirNE  *numberOfSubgridIndices];
      q_dirSW  = &subgridDistances[dirSW  *numberOfSubgridIndices];
      q_dirSE  = &subgridDistances[dirSE  *numberOfSubgridIndices];
      q_dirNW  = &subgridDistances[dirNW  *numberOfSubgridIndices];
      q_dirTE  = &subgridDistances[dirTE  *numberOfSubgridIndices];
      q_dirBW  = &subgridDistances[dirBW  *numberOfSubgridIndices];
      q_dirBE  = &subgridDistances[dirBE  *numberOfSubgridIndices];
      q_dirTW  = &subgridDistances[dirTW  *numberOfSubgridIndices];
      q_dirTN  = &subgridDistances[dirTN  *numberOfSubgridIndices];
      q_dirBS  = &subgridDistances[dirBS  *numberOfSubgridIndices];
      q_dirBN  = &subgridDistances[dirBN  *numberOfSubgridIndices];
      q_dirTS  = &subgridDistances[dirTS  *numberOfSubgridIndices];
      q_dirTNE = &subgridDistances[dirTNE *numberOfSubgridIndices];
      q_dirTSW = &subgridDistances[dirTSW *numberOfSubgridIndices];
      q_dirTSE = &subgridDistances[dirTSE *numberOfSubgridIndices];
      q_dirTNW = &subgridDistances[dirTNW *numberOfSubgridIndices];
      q_dirBNE = &subgridDistances[dirBNE *numberOfSubgridIndices];
      q_dirBSW = &subgridDistances[dirBSW *numberOfSubgridIndices];
      q_dirBSE = &subgridDistances[dirBSE *numberOfSubgridIndices];
      q_dirBNW = &subgridDistances[dirBNW *numberOfSubgridIndices];
      ////////////////////////////////////////////////////////////////////////////////
      //! - Set neighbor indices (necessary for indirect addressing)
      //!
      uint KQK = subgridDistanceIndices[k];
      uint ke   = KQK;
      uint kw   = neighborX[KQK];
      uint kn   = KQK;
      uint ks   = neighborY[KQK];
      uint kt   = KQK;
      uint kb   = neighborZ[KQK];
      uint ksw  = neighborY[kw];
      uint kne  = KQK;
      uint kse  = ks;
      uint knw  = kw;
      uint kbw  = neighborZ[kw];
      uint kte  = KQK;
      uint kbe  = kb;
      uint ktw  = kw;
      uint kbs  = neighborZ[ks];
      uint ktn  = KQK;
      uint kbn  = kb;
      uint kts  = ks;
      uint ktse = ks;
      uint kbnw = kbw;
      uint ktnw = kw;
      uint kbse = kbs;
      uint ktsw = ksw;
      uint kbne = kb;
      uint ktne = KQK;
      uint kbsw = neighborZ[ksw];
      ////////////////////////////////////////////////////////////////////////////////

      ////////////////////////////////////////////////////////////////////////////////
      //! - Set local distributions
      //!
      real f_W    = (dist.f[dirE   ])[ke   ];
      real f_E    = (dist.f[dirW   ])[kw   ];
      real f_S    = (dist.f[dirN   ])[kn   ];
      real f_N    = (dist.f[dirS   ])[ks   ];
      real f_B    = (dist.f[dirT   ])[kt   ];
      real f_T    = (dist.f[dirB   ])[kb   ];
      real f_SW   = (dist.f[dirNE  ])[kne  ];
      real f_NE   = (dist.f[dirSW  ])[ksw  ];
      real f_NW   = (dist.f[dirSE  ])[kse  ];
      real f_SE   = (dist.f[dirNW  ])[knw  ];
      real f_BW   = (dist.f[dirTE  ])[kte  ];
      real f_TE   = (dist.f[dirBW  ])[kbw  ];
      real f_TW   = (dist.f[dirBE  ])[kbe  ];
      real f_BE   = (dist.f[dirTW  ])[ktw  ];
      real f_BS   = (dist.f[dirTN  ])[ktn  ];
      real f_TN   = (dist.f[dirBS  ])[kbs  ];
      real f_TS   = (dist.f[dirBN  ])[kbn  ];
      real f_BN   = (dist.f[dirTS  ])[kts  ];
      real f_BSW  = (dist.f[dirTNE ])[ktne ];
      real f_BNE  = (dist.f[dirTSW ])[ktsw ];
      real f_BNW  = (dist.f[dirTSE ])[ktse ];
      real f_BSE  = (dist.f[dirTNW ])[ktnw ];
      real f_TSW  = (dist.f[dirBNE ])[kbne ];
      real f_TNE  = (dist.f[dirBSW ])[kbsw ];
      real f_TNW  = (dist.f[dirBSE ])[kbse ];
      real f_TSE  = (dist.f[dirBNW ])[kbnw ];
      ////////////////////////////////////////////////////////////////////////////////

      ////////////////////////////////////////////////////////////////////////////////
      //! - change the pointer to write the results in the correct array
      //!
      if (!isEvenTimestep)
      {
         dist.f[dirE   ] = &distributions[dirE   *numberOfLBnodes];
         dist.f[dirW   ] = &distributions[dirW   *numberOfLBnodes];
         dist.f[dirN   ] = &distributions[dirN   *numberOfLBnodes];
         dist.f[dirS   ] = &distributions[dirS   *numberOfLBnodes];
         dist.f[dirT   ] = &distributions[dirT   *numberOfLBnodes];
         dist.f[dirB   ] = &distributions[dirB   *numberOfLBnodes];
         dist.f[dirNE  ] = &distributions[dirNE  *numberOfLBnodes];
         dist.f[dirSW  ] = &distributions[dirSW  *numberOfLBnodes];
         dist.f[dirSE  ] = &distributions[dirSE  *numberOfLBnodes];
         dist.f[dirNW  ] = &distributions[dirNW  *numberOfLBnodes];
         dist.f[dirTE  ] = &distributions[dirTE  *numberOfLBnodes];
         dist.f[dirBW  ] = &distributions[dirBW  *numberOfLBnodes];
         dist.f[dirBE  ] = &distributions[dirBE  *numberOfLBnodes];
         dist.f[dirTW  ] = &distributions[dirTW  *numberOfLBnodes];
         dist.f[dirTN  ] = &distributions[dirTN  *numberOfLBnodes];
         dist.f[dirBS  ] = &distributions[dirBS  *numberOfLBnodes];
         dist.f[dirBN  ] = &distributions[dirBN  *numberOfLBnodes];
         dist.f[dirTS  ] = &distributions[dirTS  *numberOfLBnodes];
         dist.f[dirREST] = &distributions[dirREST*numberOfLBnodes];
         dist.f[dirTNE ] = &distributions[dirTNE *numberOfLBnodes];
         dist.f[dirTSW ] = &distributions[dirTSW *numberOfLBnodes];
         dist.f[dirTSE ] = &distributions[dirTSE *numberOfLBnodes];
         dist.f[dirTNW ] = &distributions[dirTNW *numberOfLBnodes];
         dist.f[dirBNE ] = &distributions[dirBNE *numberOfLBnodes];
         dist.f[dirBSW ] = &distributions[dirBSW *numberOfLBnodes];
         dist.f[dirBSE ] = &distributions[dirBSE *numberOfLBnodes];
         dist.f[dirBNW ] = &distributions[dirBNW *numberOfLBnodes];
      }
      else
      {
         dist.f[dirW   ] = &distributions[dirE   *numberOfLBnodes];
         dist.f[dirE   ] = &distributions[dirW   *numberOfLBnodes];
         dist.f[dirS   ] = &distributions[dirN   *numberOfLBnodes];
         dist.f[dirN   ] = &distributions[dirS   *numberOfLBnodes];
         dist.f[dirB   ] = &distributions[dirT   *numberOfLBnodes];
         dist.f[dirT   ] = &distributions[dirB   *numberOfLBnodes];
         dist.f[dirSW  ] = &distributions[dirNE  *numberOfLBnodes];
         dist.f[dirNE  ] = &distributions[dirSW  *numberOfLBnodes];
         dist.f[dirNW  ] = &distributions[dirSE  *numberOfLBnodes];
         dist.f[dirSE  ] = &distributions[dirNW  *numberOfLBnodes];
         dist.f[dirBW  ] = &distributions[dirTE  *numberOfLBnodes];
         dist.f[dirTE  ] = &distributions[dirBW  *numberOfLBnodes];
         dist.f[dirTW  ] = &distributions[dirBE  *numberOfLBnodes];
         dist.f[dirBE  ] = &distributions[dirTW  *numberOfLBnodes];
         dist.f[dirBS  ] = &distributions[dirTN  *numberOfLBnodes];
         dist.f[dirTN  ] = &distributions[dirBS  *numberOfLBnodes];
         dist.f[dirTS  ] = &distributions[dirBN  *numberOfLBnodes];
         dist.f[dirBN  ] = &distributions[dirTS  *numberOfLBnodes];
         dist.f[dirREST] = &distributions[dirREST*numberOfLBnodes];
         dist.f[dirTNE ] = &distributions[dirBSW *numberOfLBnodes];
         dist.f[dirTSW ] = &distributions[dirBNE *numberOfLBnodes];
         dist.f[dirTSE ] = &distributions[dirBNW *numberOfLBnodes];
         dist.f[dirTNW ] = &distributions[dirBSE *numberOfLBnodes];
         dist.f[dirBNE ] = &distributions[dirTSW *numberOfLBnodes];
         dist.f[dirBSW ] = &distributions[dirTNE *numberOfLBnodes];
         dist.f[dirBSE ] = &distributions[dirTNW *numberOfLBnodes];
         dist.f[dirBNW ] = &distributions[dirTSE *numberOfLBnodes];
      }
      ////////////////////////////////////////////////////////////////////////////////
      //! - rewrite distributions if there is a sub-grid distance (q) in same direction
      real q;
      q = q_dirE[k];     if (q>=c0o1 && q<=c1o1)    (dist.f[dirW  ])[kw  ]=f_E   + c4o9  * (-VeloX);
      q = q_dirW[k];     if (q>=c0o1 && q<=c1o1)    (dist.f[dirE  ])[ke  ]=f_W   + c4o9  * ( VeloX);
      q = q_dirN[k];     if (q>=c0o1 && q<=c1o1)    (dist.f[dirS  ])[ks  ]=f_N   + c4o9  * (-VeloY);
      q = q_dirS[k];     if (q>=c0o1 && q<=c1o1)    (dist.f[dirN  ])[kn  ]=f_S   + c4o9  * ( VeloY);
      q = q_dirT[k];     if (q>=c0o1 && q<=c1o1)    (dist.f[dirB  ])[kb  ]=f_T   + c4o9  * (-VeloZ);
      q = q_dirB[k];     if (q>=c0o1 && q<=c1o1)    (dist.f[dirT  ])[kt  ]=f_B   + c4o9  * ( VeloZ);
      q = q_dirNE[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirSW ])[ksw ]=f_NE  + c1o9  * (-VeloX - VeloY);
      q = q_dirSW[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirNE ])[kne ]=f_SW  + c1o9  * ( VeloX + VeloY);
      q = q_dirSE[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirNW ])[knw ]=f_SE  + c1o9  * (-VeloX + VeloY);
      q = q_dirNW[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirSE ])[kse ]=f_NW  + c1o9  * ( VeloX - VeloY);
      q = q_dirTE[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirBW ])[kbw ]=f_TE  + c1o9  * (-VeloX - VeloZ);
      q = q_dirBW[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirTE ])[kte ]=f_BW  + c1o9  * ( VeloX + VeloZ);
      q = q_dirBE[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirTW ])[ktw ]=f_BE  + c1o9  * (-VeloX + VeloZ);
      q = q_dirTW[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirBE ])[kbe ]=f_TW  + c1o9  * ( VeloX - VeloZ);
      q = q_dirTN[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirBS ])[kbs ]=f_TN  + c1o9  * (-VeloY - VeloZ);
      q = q_dirBS[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirTN ])[ktn ]=f_BS  + c1o9  * ( VeloY + VeloZ);
      q = q_dirBN[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirTS ])[kts ]=f_BN  + c1o9  * (-VeloY + VeloZ);
      q = q_dirTS[k];    if (q>=c0o1 && q<=c1o1)    (dist.f[dirBN ])[kbn ]=f_TS  + c1o9  * ( VeloY - VeloZ);
      q = q_dirTNE[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirBSW])[kbsw]=f_TNE + c1o36 * (-VeloX - VeloY - VeloZ);
      q = q_dirBSW[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirTNE])[ktne]=f_BSW + c1o36 * ( VeloX + VeloY + VeloZ);
      q = q_dirBNE[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirTSW])[ktsw]=f_BNE + c1o36 * (-VeloX - VeloY + VeloZ);
      q = q_dirTSW[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirBNE])[kbne]=f_TSW + c1o36 * ( VeloX + VeloY - VeloZ);
      q = q_dirTSE[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirBNW])[kbnw]=f_TSE + c1o36 * (-VeloX + VeloY - VeloZ);
      q = q_dirBNW[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirTSE])[ktse]=f_BNW + c1o36 * ( VeloX - VeloY + VeloZ);
      q = q_dirBSE[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirTNW])[ktnw]=f_BSE + c1o36 * (-VeloX + VeloY + VeloZ);
      q = q_dirTNW[k];   if (q>=c0o1 && q<=c1o1)    (dist.f[dirBSE])[kbse]=f_TNW + c1o36 * ( VeloX - VeloY - VeloZ);
   }
}



